#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include "module.h"
#include <arpa/inet.h>
#include <pthread.h>

#define AUTHOR "Certus" 
#define VERSION "1.5"

/*************************************************************************/
/******************** S T A R T   O F   C O N F I G - B L O C K *******************/
/*************************************************************************/

/* IF YOU DEFINE USE_ONLINECHECK WE'LL CHECK FOR ONLINE UPDATES EVERY 24H */
/* NOTE: IF YOU'VE COMPILED WITH THREADING SUPPORT, WE'LL USE A NEW THREAD */

#define USE_ONLINECHECK

/* PLEASE DEFINE THE MODULE MASTER HERE !!!!! */
/* IF YOU DON'T WE'LL MEMO ANY OF YOUR ROOTS */

// #define MODULEMASTER "Certus"

/* JUST REMOVE THE //'s AND EDIT THE NICK */

/*************************************************************************/
/********************** E N D   O F   C O N F I G - B L O C K ********************/
/*************************************************************************/

#define HTTP_PORT 80

#ifdef IRC_ULTIMATE3
# define AV_HOST_OFFSET 6
#elif defined(IRC_BAHAMUT)
# define AV_HOST_OFFSET 5
#else
# define AV_HOST_OFFSET 4
#endif

int proxy_blacklist(char *source, int ac, char **av);
char *proxy_exception_list[50];
int proxy_exceptions = 0;
int proxy_set(User *u);
int logproxy1 = 0;
int logproxy2 = 0;
void my_proxy_help(User *u);
int my_proxy_help_help(User *u);
static void *getonlineversion(void *arg);
static int check_online(int ac, char **av);
void stripspaces(char **str);
char *substr(const char *buffer, int start, int len);
void my_memo_send(char *name, char *text);
static MemoInfo *getmemoinfo(const char *name, int *ischan);

int AnopeInit(int argc, char **argv) {
	Message *m1 = NULL;
	Command *c1 = NULL;
#ifndef IRC_ULTIMATE3
	m1 = createMessage("NICK", proxy_blacklist);
#else
	m1 = createMessage("CLIENT", proxy_blacklist); /* feck you ultimate, why can't you just use "NICK" ?!?!? */
#endif
	c1 = createCommand("PROXY",proxy_set,is_services_admin,-1,-1,-1,-1,-1);
	moduleAddMessage(m1, MOD_TAIL);
	moduleAddCommand(OPERSERV,c1,MOD_HEAD);
	moduleAddAuthor(AUTHOR);
	moduleAddVersion(VERSION);
	moduleSetOperHelp(my_proxy_help);
	moduleAddHelp(c1,my_proxy_help_help);
	alog("[proxy_blacklist] proxy_blacklist.so has been loaded");
#ifndef MODULEMASTER
	alog("[proxy_blacklist] no modulemaster defined - please modify the config block in the sources of this module!");
#endif
#ifdef USE_ONLINECHECK
	moduleAddCallback("timer",time(NULL)+dotime("15s"), check_online,0,NULL);
#endif
	return MOD_CONT;
}

void AnopeFini(void) {
	int i;
	for (i=0; i <= proxy_exceptions; i++)
	{
		free(proxy_exception_list[i]);
	}
	alog("[proxy_blacklist] proxy_blacklist.so has been unloaded");
}

int proxy_blacklist(char *source, int ac, char **av) {
  struct in_addr addr;
  struct in_addr *ia;
  struct hostent *hostinfo = NULL; /* DNS result of the numeric host (DNSBL) */
  struct hostent *myalpha = NULL; /* DNS result of the alphanumeric host */
  char *list = "opm.blitzed.org"; /* our blacklist (DNSBL) */
  char *ptr = NULL, *last_octet = NULL; /* host provided my the DNSBL, its last octet which also provides the proxies (as a string) */
  char *h = NULL, *h1 = NULL, *h2 = NULL, *h3 = NULL, *h4 = NULL, *my = NULL, *last_alpha = NULL; /* host, its octets, token 
		of the alphanumeric host */
  int result; /* numerical result of proxies the users uses (== last_octet) */
  char *a = NULL; /* new, numeric host after resolving the alphanumeric host */
  int usealpha = 0; /* did we need to resolve the host because it was alphanumeric? */
  int i; /* loop integer */
  size_t size = 0; /* string sizes */
  char *my_alpha = NULL; /* test var: needed for token checks with the host */ 
  char *my_proxies = NULL; /* what kind of proxies does the user use */
  char *my_reason = NULL;
  char *test1 = "Wingate ";
  char *test2 = "SOCKS ";
  char *test3 = "HTTP-Connect ";
  char *test4 = "Router(open) ";
  char *test5 = "HTTP-Post ";

#ifndef IRC_ULTIMATE3
  if (*source) /* Is this just a nick change? return if so */
	return MOD_CONT;
#endif

/* let's check for an exception */
if (proxy_exceptions > 0) {
  for (i=0; i <= proxy_exceptions; i++)
  {
	  if (proxy_exception_list[i] == NULL)
		continue;
	  if (match_wild_nocase(proxy_exception_list[i], av[AV_HOST_OFFSET]))
	  {
		  if (logproxy1)
			alog("[proxy_blacklist] Connecting host matches exception [%s matches %s]", av[AV_HOST_OFFSET], proxy_exception_list[i]);
		  return MOD_CONT;
	  }
  }
}
  h = my = av[AV_HOST_OFFSET];

/* now we get the last token after a dot and we check whether the first letter is a digit. if not, resolve the host
   ie. "pD920B4D3.dip.t-dialin.net" needs to get resolved, 217.235.42.84 is fine */

if (av[AV_HOST_OFFSET] == NULL) {
	alog("[proxy_blacklist] internal error: av[%d] doesn't have a value", AV_HOST_OFFSET);
	return MOD_CONT;
}

if (myStrGetTokenRemainder(my, ':', 1) != NULL) {
	alog("[proxy_blacklist] Returning because of IPv6");
	return MOD_CONT;
}

if (myStrGetTokenRemainder(my, '.', 1) == NULL) {
	alog("[proxy_blacklist] Returning because of a non-dotted IP. localhost?");
	return MOD_CONT;
}

  for (i=0; i <= 8; i++) {
	my_alpha = myStrGetToken(my, '.', i);
	if (my_alpha == NULL)
	{
		last_alpha = sstrdup(myStrGetToken(my, '.', i-1));
		break;
	}
  }

if (last_alpha == NULL) {
	alog("[proxy_blacklist] internal error");
	return MOD_CONT;
}

  if (!(isdigit(*last_alpha)))
  {
	if (logproxy2)
		alog("[proxy_blacklist] Resolving alphanumeric host of %s [%s]", av[0], av[AV_HOST_OFFSET]);
	myalpha = gethostbyname(av[AV_HOST_OFFSET]);  /* myalpha is just like hostinfo */
	if (myalpha == NULL) {
		if (logproxy2)
			alog("[proxy_blacklist] Error resolving alphanumeric host of %s [%s]", av[0], av[AV_HOST_OFFSET]);
	return MOD_CONT;
	}

	ia = (struct in_addr *)(myalpha->h_addr_list[0]);
	a = inet_ntoa(*ia);
	usealpha = 1;
   }

	 /* h = ip if non-alpha; a = ip if alpha */

  h1 = myStrGetOnlyToken((usealpha) ? a : h, '.', 0);
  h2 = myStrGetToken((usealpha) ? a : h, '.', 1);
  h3 = myStrGetToken((usealpha) ? a : h, '.', 2);
  h4 = myStrGetToken((usealpha) ? a : h, '.', 3);

  size = strlen(list) + 15 + 1 + 1; /* length(list) + length (IP) + '.' + '\0' */
  h = malloc(size * sizeof(char));
  snprintf(h, size, "%s.%s.%s.%s.%s",h4,h3,h2,h1,list);
  if (logproxy2)
	alog("[proxy_blacklist] Checking numeric host of %s [%s] in the DNSBL", av[0], (usealpha) ? a : av[AV_HOST_OFFSET]);
  hostinfo = gethostbyname(h);
  free(h);
  if (hostinfo == NULL) {
	  if (logproxy2)
		 alog("[proxy_blacklist] %s's host seems to be clean [%s]", av[0], (usealpha) ? a : av[AV_HOST_OFFSET]);
	  return MOD_CONT;
  } else {
    ptr = hostinfo->h_addr_list[0];
    while(*ptr) {
      addr.s_addr = *((uint32_t *) ptr); 
      last_octet = (char *)inet_ntoa(addr);
      ptr += hostinfo->h_length;
    }
    if (last_octet) {
	  last_octet = myStrGetToken(last_octet, '.', 3); /* we wanna get the last octet X, ie. 127.1.0.X */
	  result = atoi(last_octet);


  my_proxies = NULL;
  my_proxies = malloc(51);

    sprintf(my_proxies," ");

		if (result & 1) {
			strcat(my_proxies,test1);
		}
		if (result & 2) {
			strcat(my_proxies,test2);
		}
		if (result & 4) {
			strcat(my_proxies,test3);
		}
		if (result & 8) {
			strcat(my_proxies,test4);
		}
		if (result & 16) {
			strcat(my_proxies,test5);
		}
	my_reason = malloc(450); /* this value used to be 150, but some pple have _really_ crazy hosts */
	snprintf(my_reason, 450, " Your host is listed as an open proxy in the DNSBL. Check http://opm.blitzed.org/proxy?ip=%s for more information.", av[AV_HOST_OFFSET]);
	if (logproxy1)
		alog("[proxy_blacklist] Banning %s [%s] because of the following proxies: %s", av[0], av[AV_HOST_OFFSET], my_proxies);
#if defined(IRC_UNREAL)
	send_cmd(NULL, "TKL + G %s %s %s %ld %ld :%s", "*", av[4], s_OperServ, time(NULL) + 86400 * 2, time(NULL), my_reason);
#else
	s_akill("*", av[AV_HOST_OFFSET], s_OperServ, time(NULL), time(NULL) + 86400 * 2, my_reason);
#endif
	free(my_proxies);
	free(my_reason);
	}
    return MOD_CONT;
  }
  return MOD_CONT;
}

int proxy_set(User *u)
{
	char *cmd = moduleGetLastBuffer();
	char dilim = ' ';
	char *what1 = NULL;
	char *what2 = NULL;
	char *option = NULL;
	int i = 0;


	if (cmd == NULL)
	{
		notice(s_OperServ, u->nick, "Invalid syntax: /msg operserv proxy <option> <value>");
		return MOD_CONT;
	}

	option = myStrGetToken(cmd, dilim, 0);
	what1 = myStrGetToken(cmd, dilim, 1);
	what2 = myStrGetToken(cmd, dilim, 2);

	if (!stricmp(option,"LOGPROXY")) {
		if (!what1) {
			notice(s_OperServ, u->nick, "Invalid Syntax: /msg operserv proxy logproxy <value>");
			return MOD_CONT;
		}
		if (!stricmp(what1,"0")) {
			logproxy1 = 0;
			logproxy2 = 0;
		} else if (!stricmp(what1,"1")) {
			logproxy1 = 1;
			logproxy2 = 0;
		} else if (!stricmp(what1,"2")) {
			logproxy1 = 1;
			logproxy2 = 1;
		} else {
			notice(s_OperServ, u->nick, "Invalid argument. Valid arguments are 0, 1 and 2.");
			return MOD_CONT;
		}
		notice(s_OperServ, u->nick, "Now showing Proxy messages on level %s", what1);
		alog("[proxy_blacklist] Now showing Proxy messages on level %s [set by %s]", what1, u->nick);
		return MOD_CONT;
	} else if (!stricmp(option,"EXCEPTION")) {
		if (!what1) {
			notice(s_OperServ, u->nick, "Invalid Syntax: /msg operserv proxy exception <option> [value]");
			return MOD_CONT;
		}
		if (!stricmp(what1, "ADD")) {
			if (!what2) {
				notice(s_OperServ, u->nick, "Invalid Syntax: /msg operserv proxy exception <option> [value]");
				return MOD_CONT;
			}
			if (proxy_exceptions <= 49) {
				proxy_exception_list[proxy_exceptions] = sstrdup(what2);
				++proxy_exceptions;
				notice(s_OperServ, u->nick, "Exception added for %s", what2);
				alog("[proxy_blacklist] Exception added for %s by %s", what2, u->nick);
			} else
				notice(s_OperServ, u->nick, "Proxy scanning exception list is full.");

		} else if (!stricmp(what1,"LIST")) {
			if (proxy_exception_list[0] == NULL) {
				notice(s_OperServ, u->nick, "Exception list empty");
				return MOD_CONT;
			} else {
				for (i=0; i < proxy_exceptions; i++) {
					notice(s_OperServ, u->nick, "Exception %d: %s", (i+1), proxy_exception_list[i]);
				}
			}
		}
	} else {
		notice(s_OperServ, u->nick, "Invalid option. See /msg operserv help proxy for further infos.");
		return MOD_CONT;
	}
	return MOD_CONT;
}


void my_proxy_help(User *u) {
	if (is_services_admin(u))
		notice(s_OperServ,u->nick, " PROXY       Change settings for proxy blacklist scanning");
}

int my_proxy_help_help(User *u) {
	notice(s_OperServ,u->nick,"This command allows you to change several settings for the proxy blacklist scanning.");
	notice(s_OperServ,u->nick," LOGPROXY <0|1|2>                     - Changes the level of message-reporting");
	notice(s_OperServ,u->nick," EXCEPTION ADD | LIST [pattern]  - Adds an exception for the given host pattern (wildcards allowed)");
	notice(s_OperServ,u->nick,"                                                       or shows you the current \"whitelist\" (exceptions)");
	return MOD_CONT;
}


static void *getonlineversion(void *arg)
{
	int mysocket; /* socket creation successful? */
	struct sockaddr_in host_addr;
	struct hostent *hostinfo;
	char *myhost= "modules.anope.org", *myfile = "/modules.php?op=details&id=56";
	char mycommand[512];
	char mybuf[10200];
	unsigned int bytes_sent, bytes_recv;
	char *string = mybuf;
	char *string2 = NULL, *myv = NULL;
#ifndef MODULEMASTER
	int i = 0;
#endif
	char *memotext = "[proxy_blacklist] A new version of this module has been released, check out http://modules.anope.org/modules.php?op=details&id=56";

	/* create socket */
	mysocket = socket(AF_INET, SOCK_STREAM, 0);
	if (mysocket == -1) {
		perror ("socket()");
		alog("socket error");
		exit (EXIT_FAILURE);
	}
	/* socket settings */
	memset( &host_addr, 0, sizeof (host_addr));
	host_addr.sin_family = AF_INET;
	host_addr.sin_port = htons (HTTP_PORT);
	host_addr.sin_addr.s_addr = inet_addr(myhost);

	if (host_addr.sin_addr.s_addr == INADDR_NONE) { 
		/* now we resolve the host www.computerbase.de */
		hostinfo = gethostbyname (myhost);
		if (hostinfo == NULL) {
			perror ("gethostbyname()");
			alog("gethostbyname error");
			exit (EXIT_FAILURE);
		}
		memcpy((char*) &host_addr.sin_addr.s_addr, hostinfo->h_addr, hostinfo->h_length);
	} 

	/* create a connection */
	if (connect(mysocket, (struct sockaddr *) &host_addr, sizeof(struct sockaddr)) == -1) {
		perror ("connect()");
		exit (EXIT_FAILURE);
	}

	alog("[proxy_blacklist] Checking online version....");

	/* create the HTTP-GET cmd */
	snprintf (mycommand, 512, "GET %s HTTP/1.0\nHost: %s\n\n", myfile, myhost);

	/* send cmd*/
	bytes_sent = send(mysocket, mycommand, strlen(mycommand), 0);
	if (bytes_sent == -1) {
		perror ("send()");
		exit (EXIT_FAILURE);
	}

	/* receive data and give it out */
	while ((bytes_recv = recv(mysocket, mybuf, sizeof(mybuf), 0)) > 0) {
		mybuf[bytes_recv] = '\0';	/* don't forget to cut the buffer, otherwise we might have garbage in it */
		stripspaces(&string);
		string2 = strstr(string, "Version");
		if (string2) {
			myv = substr(string2, 51, 3);
			if (!myv) {
#ifdef USE_THREADS
				pthread_exit(NULL);
#endif
				return NULL;
			}
			if (stricmp(myv, VERSION)) {
#ifdef MODULEMASTER
				my_memo_send(MODULEMASTER, memotext);
#else
	            for (i = 0; i < RootNumber; i++)
					my_memo_send(ServicesRoots[i], memotext);
#endif
				alog("[proxy_blacklist] A new version has been released online.");
			}
			else
				alog("[proxy_blacklist] No new version.");
			free(myv);
		}
	}


	if (bytes_recv == -1) {
		perror ("recv()");
		exit (EXIT_FAILURE);
	}
	printf ("\n");

	close(mysocket);
	moduleAddCallback("timer2",time(NULL)+dotime("24h"), check_online,0,NULL);
	alog("[proxy_blacklist] Online version check done. Next check in 24 hours.");
#ifdef USE_THREADS
	pthread_exit(NULL);
#endif
	return MOD_CONT;
}


void stripspaces(char **str)
{
	char *tmp = (char *)malloc(strlen(*str)+1);
	char *p = tmp;
	int i=0, first = 1;
	
	strcpy(tmp, *str);

	while(1)
	{
		while (*p && (*p == ' ' || *p == '\n' || *p == '\t'))
			p++;
	
		if (!first)
			*(*str+(i++)) = ' ';
		else
			first = 0;
	
		if (!*p) break;
	
		while (*p && !(*p == ' ' || *p == '\n' || *p == '\t'))
			*(*str+(i++)) = *(p++);

		if (!*p) break;
	}
	
	if (*(*str+i-1) == ' ')
		i--;
	
	*(*str+i) = '\0';
	
	free(tmp);
	
	return;
}

char *substr(const char *buffer, int start, int len) {
	char *ret = NULL;
	int i=0;

	if (!buffer)
		return NULL;

	ret = malloc(sizeof(char *)*(len+1));
	if(ret) {
		for(i = 0; i < len ; i++) {
			ret[i]=buffer[start+i];
		}
		ret[i]='\0';
	}
	return ret;
}

static int check_online(int ac, char **av) {

#ifdef USE_THREADS
	pthread_t mythread;

	if (pthread_create(&mythread, NULL, getonlineversion, NULL) ) {
		alog("[proxy_blacklist] Error creating new thread");
		return MOD_STOP;
	}
	if (pthread_detach(mythread)) {
		alog("[proxy_blacklist] Error detaching new thread");
		return MOD_STOP;
	}
#else
	getonlineversion(NULL);
#endif
	return MOD_CONT;
}

void my_memo_send(char *name, char *text)
{
    int ischan;
    Memo *m;
    MemoInfo *mi;
	User *u = NULL;

    if (readonly) {
        return;
    } else if (!(mi = getmemoinfo(name, &ischan))) {
		return;
    } else {
        mi->memocount++;
        mi->memos = srealloc(mi->memos, sizeof(Memo) * mi->memocount);
        m = &mi->memos[mi->memocount - 1];
        strscpy(m->sender, s_OperServ, NICKMAX);
        if (mi->memocount > 1) {
            m->number = m[-1].number + 1;
            if (m->number < 1) {
                int i;
                for (i = 0; i < mi->memocount; i++)
                    mi->memos[i].number = i + 1;
            }
        } else {
            m->number = 1;
        }
        m->time = time(NULL);
        m->text = sstrdup(text);
        m->flags = MF_UNREAD;

        if (!ischan) {
            NickAlias *na;
            NickCore *nc = (findnick(name))->nc;

            if (MSNotifyAll) {
                if (nc->flags & NI_MEMO_RECEIVE) {
                    int i;

                    for (i = 0; i < nc->aliases.count; i++) {
                        na = nc->aliases.list[i];
                        if (na->u && nick_identified(na->u))
                            notice_lang(s_MemoServ, na->u,
                                        MEMO_NEW_MEMO_ARRIVED, s_OperServ,
                                        s_MemoServ, m->number);
                    }
                } else {
                    if ((u = finduser(name)) && nick_identified(u))
                        notice_lang(s_MemoServ, u, MEMO_NEW_MEMO_ARRIVED,
                                    s_OperServ, s_MemoServ, m->number);
                }               /* if (flags & MEMO_RECEIVE) */
            }                   /* if (MSNotifyAll) */
        }                       /* if (!ischan) */
    }                           /* if command is valid */
}

static MemoInfo *getmemoinfo(const char *name, int *ischan)
{
    if (*name == '#') {
        ChannelInfo *ci;
        if (ischan)
            *ischan = 1;
        ci = cs_findchan(name);
        if (ci && !(ci->flags & CI_VERBOTEN))
            return &ci->memos;
        else
            return NULL;
    } else {
        NickAlias *na;
        if (ischan)
            *ischan = 0;
        na = findnick(name);
        if (na && !(na->status & NS_VERBOTEN))
            return &na->nc->memos;
        else
            return NULL;
    }
}


